@@ -1,14 +1,22 @@
 package Data::Graph::Util;
 
-our $DATE = '2019-02-15'; # DATE
-our $VERSION = '0.006'; # VERSION
-
 use 5.010001;
 use strict;
 use warnings;
 
 use Exporter qw(import);
-our @EXPORT_OK = qw(toposort is_cyclic is_acyclic);
+
+our $AUTHORITY = 'cpan:PERLANCAR'; # AUTHORITY
+our $DATE = '2023-12-20'; # DATE
+our $DIST = 'Data-Graph-Util'; # DIST
+our $VERSION = '0.007'; # VERSION
+
+our @EXPORT_OK = qw(
+                       toposort
+                       is_cyclic
+                       is_acyclic
+                       connected_components
+               );
 
 sub _toposort {
     my $graph = shift;
@@ -72,6 +80,46 @@ sub is_acyclic {
     !$err;
 }
 
+sub connected_components {
+    my $graph = shift;
+
+    # create a map of bidirectional connections between nodes, to ease checking
+    my %connections;
+    for my $node1 (keys %$graph) {
+        for my $node2 (@{ $graph->{$node1} }) {
+            $connections{$node1}{$node2} = 1;
+            $connections{$node2}{$node1} = 1;
+        }
+    }
+
+    my @subgraphs;
+    my %remaining_nodes = %$graph;
+
+    # traverse a node to get a subgraph. remove the nodes from the original
+    # graph. repeat until there are no nodes left on the original graph.
+
+    while (1) { # while there are still unlabeled nodes
+        my ($node1, $dependants1) = each %remaining_nodes or last;
+
+        my $subgraph = {$node1 => $dependants1};
+        my %seen;
+        my @nodes_to_check = keys %{ $connections{$node1} };
+
+        while (@nodes_to_check) { # while we can still find nodes connected to the subgraph
+            my $node2 = shift @nodes_to_check;
+            next if $seen{$node2}++;
+            if (my $dependants2 = delete $remaining_nodes{$node2}) {
+                $subgraph->{$node2} = $dependants2;
+                push @nodes_to_check, keys %{ $connections{$node2} };
+            }
+        }
+
+        push @subgraphs, $subgraph;
+    }
+
+    sort { scalar(keys %$b) <=> scalar(keys %$a) } @subgraphs;
+}
+
 1;
 # ABSTRACT: Utilities related to graph data structure
 
@@ -87,20 +135,25 @@ Data::Graph::Util - Utilities related to graph data structure
 
 =head1 VERSION
 
-This document describes version 0.006 of Data::Graph::Util (from Perl distribution Data-Graph-Util), released on 2019-02-15.
+This document describes version 0.007 of Data::Graph::Util (from Perl distribution Data-Graph-Util), released on 2023-12-20.
 
 =head1 SYNOPSIS
 
- use Data::Graph::Util qw(toposort is_cyclic is_acyclic);
+ use Data::Graph::Util qw(
+     toposort
+     is_cyclic
+     is_acyclic
+     connected_components
+ );
 
- # return nodes that satisfy the following graph: a must come before b, b must
- # come before c & d, and d must come before c.
+ # return nodes of a graph. a must come before b, b must come before c & d, and
+ # d must come before c.
 
  my @sorted = toposort(
      { a=>["b"], b=>["c", "d"], d=>["c"] },
  ); # => ("a", "b", "d", "c")
 
- # sort specified nodes (2nd argument) using the graph. nodes not mentioned in
+ # sort nodes specified in 2nd argument using the graph. nodes not mentioned in
  # the graph will be put at the end. duplicates are not removed.
 
  my @sorted = toposort(
@@ -118,10 +171,26 @@ This document describes version 0.006 of Data::Graph::Util (from Perl distributi
  say is_cyclic ({a=>["b"], b=>["c"], c=>["a"]}); # => 1
  say is_acyclic({a=>["b"], b=>["c"], c=>["a"]}); # => 0
 
+ # return connected subgraphs, sorted by the largest
+ my @subgraphs = connected_components(
+     { a=>["b"], b=>["c", "d"], d=>["c"] },
+ ); # => return 1 element, all nodes are connected as a single graph
+
+ # return connected subgraphs, sorted by the largest
+ my @subgraphs = connected_components(
+     { a=>["b"], b=>["c", "d"], d=>["c"],
+       e=>["f"],
+       g=>["h", "i"], j=>["a"],
+     },
+ ); # => return 3 elements, there are 3 separate subgraphs
+    # ({a=>["b"], b=>["c", "d"], d=>["c"], j=>["a"]}, {e=>["f"]}, {g=>["h","i"]})
+
 =head1 DESCRIPTION
 
 Early release. More functions will be added later.
 
+This module provides some functions related to the graph data structure.
+
 Keywords: topological ordering, dependency sorting, dependency ordering.
 
 =head1 FUNCTIONS
@@ -161,6 +230,16 @@ Usage:
 Return true if graph is acyclic, i.e. contains no cycles. The opposite of
 L</is_cyclic>.
 
+=head2 connected_components
+
+Usage:
+
+ connected_components(\%graph) => list of subgraphs
+
+Return list of subgraphs that are not connected to one another, sorted by
+descending size. If all the nodes are connected, will return a single subgraph
+(the original graph itself).
+
 =head1 HOMEPAGE
 
 Please visit the project's homepage at L<https://metacpan.org/release/Data-Graph-Util>.
@@ -169,20 +248,20 @@ Please visit the project's homepage at L<https://metacpan.org/release/Data-Graph
 
 Source repository is at L<https://github.com/perlancar/perl-Data-Graph-Util>.
 
-=head1 BUGS
-
-Please report any bugs or feature requests on the bugtracker website L<https://rt.cpan.org/Public/Dist/Display.html?Name=Data-Graph-Util>
-
-When submitting a bug or request, please include a test-file or a
-patch to an existing test-file that illustrates the bug or desired
-feature.
-
 =head1 SEE ALSO
 
+=head2 Articles
+
 L<https://en.wikipedia.org/wiki/Graph_(abstract_data_type)>
 
 L<https://en.wikipedia.org/wiki/Topological_sorting#Kahn.27s_algorithm>
 
+=head2 Related modules
+
+L<Graph> contains more graph-related algorithms.
+
+=head3 Topological sort
+
 L<Algorithm::Dependency> can also do topological sorting, but it is more finicky
 with input: graph cannot be epmty and all nodes need to be specified.
 
@@ -195,11 +274,43 @@ See L<Bencher::Scenario::GraphTopologicalSortModules> for benchmarks.
 
 perlancar <perlancar@cpan.org>
 
+=head1 CONTRIBUTOR
+
+=for stopwords Jose Luis Martínez Torres
+
+Jose Luis Martínez Torres <joseluis.martinez@capside.com>
+
+=head1 CONTRIBUTING
+
+
+To contribute, you can send patches by email/via RT, or send pull requests on
+GitHub.
+
+Most of the time, you don't need to build the distribution yourself. You can
+simply modify the code, then test via:
+
+ % prove -l
+
+If you want to build the distribution (e.g. to try to install it locally on your
+system), you can install L<Dist::Zilla>,
+L<Dist::Zilla::PluginBundle::Author::PERLANCAR>,
+L<Pod::Weaver::PluginBundle::Author::PERLANCAR>, and sometimes one or two other
+Dist::Zilla- and/or Pod::Weaver plugins. Any additional steps required beyond
+that are considered a bug and can be reported to me.
+
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2019, 2017, 2016 by perlancar@cpan.org.
+This software is copyright (c) 2023, 2019, 2017, 2016 by perlancar <perlancar@cpan.org>.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
 
+=head1 BUGS
+
+Please report any bugs or feature requests on the bugtracker website L<https://rt.cpan.org/Public/Dist/Display.html?Name=Data-Graph-Util>
+
+When submitting a bug or request, please include a test-file or a
+patch to an existing test-file that illustrates the bug or desired
+feature.
+
 =cut
